% Predecessor: SQG_v04 created during concurrent maximization of newsvendor profit and computation of CVaR from
% continuous distribution

% Differences with predecessor:

% - possibility to have expectation functions in constraints
% - changes in the algorithm parameters that support expectation
%   constraints

% For the future

% - default computation of initial step and step scales after some initial observations of
%   the gradients, something like one hundredth of the ratio between
%   typical range of the variables (this is already in the parameters) and
%   typical norm of the gradient during specified number of initial
%   iterations
% - introduce separate averaging speeds for different variables and
%   objective, in CVaR example function averages converge more slowly than
%   VaR average. What is the optimal speed of averaging?


function [xOut, fval, exitFlag, InfIter] = SQG_v05(fun, parFun, parAlg, x0, funConstr, parConstr)

%  For description of parameters see reference manual

%  Initial assignment of output variables

xOut=[]; 
fval=[]; 
InfIter=[];
exitFlag=0;

% Initialization of parameters of the algorithm 

[illegalInput, NumProblems, nVars, ProblemNames, ProblemStates, ProblemType, ...
    DPrimary, NSecondaries, Secondaries, DSecondaries, SecondariesAveraging, IsSecondary, ...
    iter, iavg, stepScale, step0, schedule, stepmult, ProblemScale, FDiff, gamma, ...
    TraceAveraging, MaxVarChange, VarRange, PlotFunctionAvg, PlotFunctionObs, RefFunctionValue, ...
    PlotVariableNumber, RefVariablesValues, SaveFileName, RunComment, randInitFlag, ...
    ConNumber, ConTypes, ConRHS, ConRef, EstimateGradient, ConProcessing, ConPenalty, ...
    PlotConstraintsViolation, SaveFigures, KeepTrack] = SQGParameters_v03( parAlg, x0 );   %#ok<ASGLU>
if illegalInput
    exitFlag=2;
    disp('SQG exits due to illegal assignments to algorithm parameters');
    return
end

% For explanation of algorithm parameters see Reference Manual and
% correspondence of parAlg keywords to identifiers used in the algorithm

% reset of random number generator

if randInitFlag
    rng('default');
end

% 2. Initial setup

M=max(nVars);
MS=max(ProblemStates);
MC=max(ConNumber);
MF=max(NSecondaries);
simulationFlag=(MS>0);
NeedFiniteDifferences=zeros(1,NumProblems);
for i=1:NumProblems
    if ConNumber(i) > 0
        NeedFiniteDifferences(i)=max(DPrimary(i),max(DSecondaries(i,:)));
    else
        NeedFiniteDifferences(i)=DPrimary(i);
    end
end
fdiffFlag=max(NeedFiniteDifferences);  % at least one finite difference among the problems and constraints
gradFlag=max(DPrimary==0);             % at least one gradient among the problems

xOut=zeros(NumProblems,M);
fval=zeros(1,NumProblems);
ConV=zeros(NumProblems,MC);
ConVFin=zeros(NumProblems,MC);
ConVAvg=zeros(NumProblems,MC);
if KeepTrack
    bufVar=zeros(NumProblems,M,iter+1);
    bufConV=zeros(NumProblems,MC,iter);
    bufConVAvg=zeros(NumProblems,MC,iter);
    bufVar(:,:,1)=x0;
    bufVarAvg=zeros(NumProblems,M,iter+1);
    bufVarAvg(:,:,1)=x0;
    bufF=zeros(NumProblems,iter);
    bufFAvg=zeros(NumProblems,iter);
else
    bufVar=[];
    bufConV=[];
    bufConVAvg=[];
    bufVarAvg=[];
    bufF=[];
    bufFAvg=[];
end
if fdiffFlag
    fEstD=zeros(NumProblems,M,MF+1);
end
step=zeros(NumProblems,M);
BasicStep=zeros(NumProblems,M);
for i=1:NumProblems
    BasicStep(i,1:nVars(i))=step0(i)*stepScale{i};
    step(i,1:nVars(i))=ProblemScale{i}(1)*BasicStep(i,1:nVars(i));
end

StepPointer=ones(1,NumProblems);
x=x0;
xA=x0;

if simulationFlag
    State=zeros(NumProblems,MS);
    if fdiffFlag
        StateD=zeros(NumProblems,MS,M,MF+1);
    end
else
    State=[];
end

if EstimateGradient
    GradEst=zeros(NumProblems,M,MF+1);
else
    
end

% Check nonexpectation constraints data and constraints parameters

if iscell(funConstr)
    if length(funConstr) ~=NumProblems
        disp('SQG exits because the number of projection functions is not equal to the number of problems')
        exitFlag=4;
        return
    end
elseif  isa(funConstr,'function_handle')
    buf=funConstr;
    funConstr=cell(1,NumProblems);
    for i=1:NumProblems
        funConstr{i}=buf;
    end
else
    disp('SQG exits due to illegal definition of projection functions')
    exitFlag=4;
    return
end
if NumProblems == 1
    parConstr={parConstr};
end
if  ~iscell(parConstr) || (length(parConstr) ~= NumProblems)
    disp('SQG exits due to illegal structure of projection data')
    exitFlag=4;
    return
end

% Initialization of objectives

modeFlag=int32(-1);
if randInitFlag
    randFlag=int8(2);  % This is because mex does not hear initialized state from calling function
else
    randFlag=int8(1);
end
[y, grad, fatalError, State]=fun(x, parFun, modeFlag, randFlag, State);
if fatalError
    [xOut, fval, exitFlag, InfIter]=abortActions_v02(-1,0,x,y,bufVar,bufVarAvg,bufF,bufFAvg,...
        bufConV,bufConVAvg);
    return
end

% Check of correct construction of outputs of user function

Ny=size(y);
if (Ny(1) ~=1) || (Ny(2) ~= NumProblems)
    disp('SQG aborts at initialization due to wrong dimension of function vector') 
    exitFlag=3;
    return
end
Ng=size(grad);
if length(Ng) == 2  % this is for problems without expectation constraints
    Ng=[Ng 1];
end
if gradFlag
    if (Ng(1) ~= NumProblems) || (Ng(2) ~= M) || (Ng(3) ~= MC+1)
        disp('SQG aborts at initialization due to wrong dimension of gradient matrix') 
        exitFlag=3;
        return
    end
else
    if ~isempty(grad)
        disp('SQG aborts at initialization due to nonempty gradient matrix in the absense of gradients') 
        exitFlag=3;
        return
    end
end
NNs=size(State);
if simulationFlag
    if (NNs(1) ~= NumProblems) || (NNs(2) ~= MS)
        disp('SQG aborts at initialization due to wrong dimension of state matrix') 
        exitFlag=3;
        return
    end
else
    if ~isempty(State)
        disp('SQG aborts at initialization due to nonempty state matrix in the absense of states') 
        exitFlag=3;
        return
    end
end

% initialization of function estimates and states

fA=y;        % estimate for the process trace and final function value
EstSecondaries=y;    % estimate for measuring constraint violation
if EstimateGradient
    GradEst=grad; 
end
if fdiffFlag
    fEst=y;  % estimate to compute finite differences
    for i=1:NumProblems
        if DPrimary(i) 
            fEstD(i,1:nVars(i),1)=y(i);
            if ProblemStates(i) > 0
                for j=1:nVars(i)
                    StateD(i,1:ProblemStates(i),j,1)=State(i,1:ProblemStates(i));  
                end
            end        
        end       
        for j=1:NSecondaries(i)
            if DSecondaries(i,j) == 1
                fEstD(i,1:nVars(i),j+1)=y(Secondaries(i,j)); 
                if ProblemStates(Secondaries(i,j)) > 0
                    for jj=1:nVars(i)
                        StateD(i,1:ProblemStates(Secondaries(i,j)),jj,j+1)= ...
                            State(Secondaries(i,j),1:ProblemStates(Secondaries(i,j)));  
                    end
                end
            end
        end
    end
end

% SQG cycle

for i=1:iter
    avg=min(1,TraceAveraging(:,1)./(TraceAveraging(:,2)+i.^TraceAveraging(:,3)));
    AvgSecondaries=min(1,SecondariesAveraging(:,1)./(SecondariesAveraging(:,2)+i.^SecondariesAveraging(:,3)));
    modeFlag=int32(0);
    
    %  New function observation
    
    randFlag=int8(1);
    if i~=1
        [y, grad, fatalError, State]=fun(x, parFun, modeFlag, randFlag, State);
        if fatalError
            [xOut, fval, exitFlag, InfIter]=abortActions_v02(0,i,x,y,bufVar,bufVarAvg,...
                bufF,bufFAvg,bufConV,bufConVAvg);
            return
        end
        
        % update of basic functions estimate for finite differences
        
        % Actually, fEst(j) is needed only if
        % - objective of problem j does not have gradient with respect to
        %   the problem variables
        % - objective of problem j enters as constraint function to some
        % problem k and does not have gradient with respect to variables of
        % this problem
        
        if fdiffFlag
            fEst=(1-gamma).*fEst+gamma.*y;
        end
        
        % update of secondaries estimates 
        
        if MF > 0
            for j=1: NumProblems
                if IsSecondary(j)
                    EstSecondaries(j)=(1-AvgSecondaries(j))*EstSecondaries(j)+AvgSecondaries(j)*y(j);
                end
            end
        end
                   
    end
   
    %  New observations of function values at finite differences
    
    if fdiffFlag
        if isempty(grad)
            grad=zeros(NumProblems,M,MF+1);
        end
        randFlag=int8(0);
        StateBuf=State;
        for j=1:NumProblems
            if ProblemType(j) == 0
                continue
            end
            if NeedFiniteDifferences(j) 
                modeFlag=int32(j);
                for k=1:nVars(j)
                    
                    % modify variables
                    
                    x(j,k)=x(j,k)+FDiff{j}(k);
                    
                    % put relevant states in state buffer to send to user function, 
                    % i.e. for those functions of k-th
                    % problem for which the gradient does not exist
                    
                    if DPrimary(j) && (ProblemStates(j) > 0)
                        StateBuf(j,1:ProblemStates(j))=StateD(j,1:ProblemStates(j),k,1);
                    end
                    for kk=1:NSecondaries(j)
                        if DSecondaries(j,kk) && (ProblemStates(Secondaries(j,kk)) > 0)
                            StateBuf(Secondaries(j,kk),1:ProblemStates(Secondaries(j,kk)))=...
                                StateD(j,1:ProblemStates(Secondaries(j,kk)),k,kk+1);
                        end
                    end
                    
                    % call user function with k - mode
                    
                    [yy, ~, fatalError, StateBuf]=fun(x, parFun, modeFlag, randFlag, StateBuf);
                    if fatalError
                        [xOut, fval, exitFlag, InfIter]=abortActions_v02(j,i,x,y,bufVar,bufVarAvg,bufF,...
                            bufFAvg,bufConV,bufConVAvg);
                        return
                    end
                    
                    % modify back the variables
                    
                    x(j,k)=x(j,k)-FDiff{j}(k);
                    
                    % store new states
                    
                    if DPrimary(j) && ProblemStates(j) > 0
                        StateD(j,1:ProblemStates(j),k,1)=StateBuf(j,1:ProblemStates(j));
                    end
                    for kk=1:NSecondaries(j)
                        if DSecondaries(j,kk) && (ProblemStates(Secondaries(j,kk)) > 0)
                            StateD(j,1:ProblemStates(Secondaries(j,kk)),k,kk+1)=...
                                StateBuf(Secondaries(j,kk),1:ProblemStates(Secondaries(j,kk)));
                        end
                    end
                    
                    % estimate gradient of objective
                    
                    if DPrimary(j)
                        fEstD(j,k,1)=(1-gamma(j))*fEstD(j,k,1)+gamma(j)*yy(j);
                        grad(j,k,1)=(fEstD(j,k,1)-fEst(j))/FDiff{j}(k);
                    end
                    
                    % estimate gradient of constraints
                    
                    for kk=1:NSecondaries(j)
                        if DSecondaries(j,kk) 
                            fEstD(j,k,kk+1)=(1-gamma(j))*fEstD(j,k,kk+1)+gamma(j)*yy(Secondaries(j,kk)); % that is, gamma is problem specific and not function specific
                            grad(j,k,kk+1)=(fEstD(j,k,kk+1)-fEst(Secondaries(j,kk)))./FDiff{j}(k);
                        end
                    end
                end
            end
        end
    end
    
    for j=1:NumProblems    
        
        %  Update stepsize
    
        if ProblemType(j) == 0
                continue
        end
        if StepPointer(j) <= length(schedule{j}) % stepsize update
            if i == schedule{j}(StepPointer(j))
                StepPointer(j)=StepPointer(j)+1;
                BasicStep(j,1:nVars(j))=BasicStep(j,1:nVars(j))*stepmult(j);
                step(j,1:nVars(j))=BasicStep(j,1:nVars(j))*ProblemScale{j}(StepPointer(j));
            end
        end
        
        % Processing of expectation constraints
        
        if ConNumber(j) > 0
            ViolationDetected=0;
            gradComp=zeros(1,nVars(j));
            if EstimateGradient
                normGrad=sqrt(GradEst(j,1:nVars(j),1)*GradEst(j,1:nVars(j),1)');
            else
                normGrad=1;
            end
            for k=1:ConNumber(j)
                buf=EstSecondaries(Secondaries(j,ConRef(j,k)))-ConRHS(j,k);
                if ConTypes(j,k) == 1
                    if buf < 0
                        ConV(j,k)=0;
                        continue
                    else
                        if ConProcessing(j,1) == 1
                            gCon=-grad(j,1:nVars(j),ConRef(j,k)+1);
                        elseif ConProcessing(j,1) == 2
                            gCon=-buf*grad(j,1:nVars(j),ConRef(j,k)+1);
                        end                    
                        ConV(j,k)=buf;
                        ViolationDetected=1;
                    end
                elseif ConTypes(j,k) == 2
                    if buf > 0
                        ConV(j,k)=0;
                        continue
                    else
                        if ConProcessing(j,1) == 1
                            gCon=grad(j,1:nVars(j),ConRef(j,k)+1);
                        elseif ConProcessing(j,1) == 2
                            gCon=-buf*grad(j,1:nVars(j),ConRef(j,k)+1);
                        end 
                        ConV(j,k)=-buf;
                        ViolationDetected=1;
                    end
                elseif ConTypes(j,k) == 3
                    if buf == 0
                        ConV(j,k)=0;
                        continue
                    else
                        if ConProcessing(j,1) == 1
                            if buf < 0
                                gCon=grad(j,1:nVars(j),ConRef(j,k)+1);
                            else
                                gCon=-grad(j,1:nVars(j),ConRef(j,k)+1);
                            end
                        elseif ConProcessing(j,1) == 2
                            gCon=-buf*grad(j,1:nVars(j),ConRef(j,k)+1);
                        end
                        ConV(j,k)=abs(buf);
                        ViolationDetected=1;
                    end  
                end
                if ConV(j,k) > 0
                    if EstimateGradient
                        normCon=sqrt(grad(j,1:nVars(j),ConRef(j,k)+1)*grad(j,1:nVars(j),ConRef(j,k)+1)');
                    else
                        normCon=1;
                    end
                    if normCon > 0
                        Cmult=normGrad/normCon;
                    end
                    gradComp=gradComp-ProblemType(j)*Cmult*gCon;
                end
            end
            if ViolationDetected
                if ~ConProcessing(j,2) 
                    gradComp=grad(j,1:nVars(j),1)+ConPenalty(j)*gradComp;
                else
                    gradComp=ConPenalty(j)*gradComp;
                end
            else
                gradComp=grad(j,1:nVars(j),1);
            end
        else
            gradComp=grad(j,1:nVars(j),1);
        end

        %  Compute step move
        
        if ConNumber(j) == 0 
            xc = step(j,1:nVars(j)).*grad(j,1:nVars(j));
        else
            xc = step(j,1:nVars(j)).*gradComp;
        end
        rc =max(abs(xc)./VarRange{j});
        if rc > MaxVarChange(j)
            xc=xc*MaxVarChange(j)/rc;
        end
        
        % make step move
        
        x(j,1:nVars(j)) = x(j,1:nVars(j)) - ProblemType(j)*xc;

        %  Make projection and obtain averaged point

        x(j,1:nVars(j))=funConstr{j}(x(j,1:nVars(j)),parConstr{j});
        
        % average trace
        
        xA(j,1:nVars(j))=(1-avg(j))*xA(j,1:nVars(j))+avg(j)*x(j,1:nVars(j));
        if EstimateGradient
            GradEst(j,1:nVars(j),1:NSecondaries(j)+1)=(1-avg(j))* GradEst(j,1:nVars(j),1:NSecondaries(j)+1)+...
            avg(j)*grad(j,1:nVars(j),1:NSecondaries(j)+1);
        end
        
        % Keep record
        
        if KeepTrack
            bufVar(j,1:nVars(j),i+1)=x(j,1:nVars(j));
            bufVarAvg(j,1:nVars(j),i+1)=xA(j,1:nVars(j));   
            if ConNumber(j) > 0
                bufConV(j,1:ConNumber(j),i)=ConV(j,1:ConNumber(j));
                bufConVAvg(j,1:ConNumber(j),i)=ConVAvg(j,1:ConNumber(j));
            end
        end
    
    end
    
    for j=1:NumProblems
        fA(j)=(1-avg(j))*fA(j)+avg(j)*y(j);
        if ConNumber(j) > 0
            ConVAvg(j,1:ConNumber(j))=(1-avg(j))*ConVAvg(j,1:ConNumber(j))+...
                avg(j)*ConV(j,1:ConNumber(j));
        end
    end
    
    if KeepTrack
        bufF(:,i)=y';
        bufFAvg(:,i)=fA';
    end
    
end

% averaging for obtaining solution

iavg=min(iavg,iter);
for j=1:NumProblems
    if KeepTrack
        xOut(j,1:nVars(j))=mean(bufVarAvg(j,1:nVars(j),iter+2-iavg(j):iter+1),3);
        fval(j)=mean(bufFAvg(j,iter+1-iavg(j):iter));
        if ConNumber(j) > 0
            ConVFin(j,1:ConNumber(j))=mean(bufConVAvg(j,1:ConNumber(j),iter+1-iavg(j):iter),3);
        end
    else
        xOut(j,1:nVars(j))=xA(j,1:nVars(j));
        fval(j)=fA(j);
        if ConNumber(j) > 0
            ConVFin(j,1:ConNumber(j))=ConVAvg(j,1:ConNumber(j));
        end
    end
end

% Save iterations record

if KeepTrack
    InfIter = {{'variables' bufVar} {'averaged variables' bufVarAvg} ...
        {'function observations' bufF} {'averaged observations' bufFAvg} ...
        {'observation of constraints violation', bufConV} ...
        {'averaged constraints violation', bufConVAvg} {'constraint satisfaction' ConVFin }};
else
    InfIter=[];
end
if ~isempty(SaveFileName)
    save(SaveFileName,'RunComment','xOut','fval','InfIter','fun','parFun','parAlg',...
        'x0','funConstr','parConstr');
end

% Plot process trace

if KeepTrack
    PlotSQGTrace_v02( NumProblems, ProblemNames, PlotFunctionAvg, PlotFunctionObs, RefFunctionValue, ...
        PlotVariableNumber, RefVariablesValues, PlotConstraintsViolation, iter, bufF, bufFAvg, ...
        bufVar, bufVarAvg, ConNumber, bufConV, bufConVAvg, ConRef, Secondaries, SaveFigures )  
end
    
end


